<!DOCTYPE html>
<html>
<head>

</head>

<body>

<title>Transphorm's NuGet Server</title>
<h1>Transphorm's NuGet Server</h1>

<div class=page>
  <div id="content">
    {% block body %}
    {% endblock %}
  </div>
</div>

</body>
</html>